from __future__ import annotations

from typing import List, Union, Type, TYPE_CHECKING

import types
from abc import ABC, abstractmethod
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.testresult import FixturePartResult, ResultState, ResultSummary, TestcaseResult

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.scenario import Scenario


class BasicExecutor(ABC):
    """
    The BasicExecutor class is an abstract class that represents the parent class of all executors. Together with other
    executor classes, an executor forms a tree structure in which individual tests, which later on are executed, are
    assigned to individual scenarios
    """
    # this property describes the runnable state of the executor branch before the executor is really used
    #  with this you can declare a whole branch as inactive, while the collecting process is active
    prev_mark = PreviousExecutorMark.RUNNABLE

    def __init__(self):
        # contains the result object for the CONSTRUCT FIXTURE part of this branch
        self.construct_result = FixturePartResult(self)
        # contains the result object for the BODY part of this branch (will be defined in subclasses)
        self.body_result = None
        # contains the result object for the TEARDOWN FIXTURE part of this branch
        self.teardown_result = FixturePartResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    @abstractmethod
    def all_child_executors(self) -> list[BasicExecutor] | None:
        """
        returns all child executors of this object or None if no child executors can exist (this element is a leaf)
        """

    @property
    @abstractmethod
    def parent_executor(self) -> BasicExecutor:
        """returns the parent executor of this element"""

    @property
    @abstractmethod
    def base_instance(self) -> object:
        """returns the base class instance to which this executor instance belongs or None if this element is a
        ExecutorTree"""

    @property
    def executor_result(self) -> ResultState:
        """
        This property returns the combined state of this executor object. This is determined by its properties
        ``construct_result``, ``body_result`` and ``teardown_result``.
        """
        relative_result = ResultState.NOT_RUN
        priority_order = ResultState.priority_order()

        if priority_order.index(self.construct_result.result) < priority_order.index(relative_result):
            relative_result = self.construct_result.result
        if priority_order.index(self.body_result.result) < priority_order.index(relative_result):
            relative_result = self.body_result.result
        if priority_order.index(self.teardown_result.result) < priority_order.index(relative_result):
            relative_result = self.teardown_result.result
        return relative_result

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_all_recognized_exception(self) -> List[Exception]:
        """
        This method returns all occurred Exception of this branch.

        :return: a list with all occurred and recorded exceptions
        """
        all_own = [self.construct_result.exception, self.body_result.exception, self.teardown_result.exception]
        if self.all_child_executors:
            for cur_child_executor in self.all_child_executors:
                all_own += cur_child_executor.get_all_recognized_exception()

        # filter all duplicated entries
        all_own = list(set(all_own))
        if None in all_own:
            all_own.remove(None)
        return all_own

    def has_runnable_tests(self, consider_discarded_too=False) -> bool:
        """
        This method returns true if this executor element is runnable. The method returns true if this element has
        `prev_mark=RUNNABLE` and minimum one of its children has `prev_mark=RUNNABLE` too.

        :param consider_discarded_too: True if the method allows DISCARDED elements too
        """
        allowed_prev_marks = [PreviousExecutorMark.RUNNABLE]

        if consider_discarded_too:
            allowed_prev_marks.append(PreviousExecutorMark.DISCARDED)

        if self.prev_mark not in allowed_prev_marks:
            return False

        if self.all_child_executors is not None:
            # the executor has child executors -> check them
            for cur_child in self.all_child_executors:
                if cur_child.has_runnable_tests(consider_discarded_too):
                    return True
            return False
        return True

    def has_skipped_tests(self) -> bool:
        """
        This method returns true if this executor element has at least one test that is marked to be skipped. The method
        returns true if minimum one of its children has `prev_mark=SKIP`.
        """
        if self.all_child_executors is not None:
            # the executor has child executors -> check them
            for cur_child in self.all_child_executors:
                if cur_child.has_skipped_tests():
                    return True
            return False
        return False

    def has_covered_by_tests(self) -> bool:
        """
        This method returns true if this executor element has at least one test that is marked as covered-by. The method
        returns true if minimum one of its children has `prev_mark=COVERED_BY`.
        """
        if self.all_child_executors is not None:
            # the executor has child executors -> check them
            for cur_child in self.all_child_executors:
                if cur_child.has_covered_by_tests():
                    return True
            return False
        return False

    def get_all_base_instances_of_this_branch(
            self, with_type: Union[Type[Setup], Type[Scenario], Type[types.FunctionType]],
            only_runnable_elements: bool = True) -> List[Union[Setup, Scenario, object]]:
        """
        This method returns a list of all base instance elements that are from given type `with_type`. If `with_type`
        is a base type of this branch element, the method always returns a list with one element. If the base instance
        element is a child of the current branch, it could be a list with zero to infinite elements.

        .. note::
            The method only returns unique objects.
        """
        # search all higher classes and this one (if there is a match, return value is a list with exactly one element)
        cur_executor = self
        # only go through cur_executor == ExecutorTree (do not iterate with this object)
        while cur_executor.parent_executor is not None:
            if isinstance(cur_executor.base_instance, with_type):
                if not only_runnable_elements or cur_executor.has_runnable_tests():
                    return [cur_executor.base_instance]
            cur_executor = cur_executor.parent_executor

        # go in the branch and search in all children (recursively)
        result = []
        # if there are no child executors -> we do not need to search in children because there are no children
        if self.all_child_executors is not None:
            for cur_child_executor in self.all_child_executors:
                child_result = cur_child_executor.get_all_base_instances_of_this_branch(
                    with_type=with_type, only_runnable_elements=only_runnable_elements)
                result += child_result
        # remove duplicate items
        return list(set(result))

    @abstractmethod
    def cleanup_empty_executor_branches(self, consider_discarded=False):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.

        :param consider_discarded: true if this method should consider discarded branches, otherwise False
        """

    def filter_tree_for_user_filters(self):
        """
        This method calls all user defined filters that are to be applied to the executor tree.
        """
        if self.all_child_executors:
            for cur_child_executor in self.all_child_executors:
                cur_child_executor.filter_tree_for_user_filters()

    def testsummary(self) -> ResultSummary:
        """
        returns a dictionary with all possible :class:`ResultState` as keys and the number of times they have occurred
        in this branch as values
        """

        summary = ResultSummary()

        if isinstance(self.body_result, TestcaseResult):
            setattr(summary, self.executor_result.value, 1)
        elif self.all_child_executors:
            for cur_child_exec in self.all_child_executors:
                summary += cur_child_exec.testsummary()
        return summary



import balder


class FeatureI(balder.Feature):

    def do_something(self):
        raise NotImplementedError("has to be overwritten in subclass")


class FeatureII(balder.Feature):

    def do_something(self):
        raise NotImplementedError("has to be overwritten in subclass")


class FeatureIII(balder.Feature):

    def do_something(self):
        raise NotImplementedError("has to be overwritten in subclass")


class FeatureIV(balder.Feature):

    def do_something(self):
        raise NotImplementedError("has to be overwritten in subclass")


from __future__ import annotations

import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__file__)


class Controller(ABC):
    """
    This is the base controller class. It serves as a base class for all other specialized controllers.

    A controller is used to manage the behavior internally.
    """

    @staticmethod
    @abstractmethod
    def get_for(related_cls):
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """

    @property
    @abstractmethod
    def related_cls(self):
        """
        This method returns the related class that belongs to that controller.
        """



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_calculator(balder_working_dir):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except ValueError as exc:
        assert exc.args[0] == ("the argument `not_existing` does not exist in test method "
                               "`ScenarioAdding.test_add_two_numbers`")

    assert session.executor_tree is None, "test session does not terminates before collector work was done"




from ...lib.features import FeatureIII
from ...balderglob import RuntimeObserver


class SetupFeatureIII(FeatureIII):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIII, SetupFeatureIII.do_something, "enter `SetupFeatureIII.do_something`",
            category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIII, SetupFeatureIII.called_from_outer_feature,
            "enter `SetupFeatureIII.called_from_outer_feature`", category="feature")



from enum import Enum


class ExitCode(Enum):
    """
    This enum describes the balder exit codes.
    """
    #: tests were collected and all tests and fixtures terminates with success
    SUCCESS = 0
    #: tests were collected but some tests failed
    TESTS_FAILED = 1
    #: test was interrupted by user
    USER_INTERRUPT = 2
    #: an internal unexpected error occurs
    UNEXPECTED_ERROR = 3
    #: Balder usage error
    BALDER_USAGE_ERROR = 4
    #: no tests were collected
    NO_TESTS_COLLECTED = 5



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_calculator(balder_working_dir):
    """
    This testcase executes the basic example but with two Scenarios. The parent scenario adds the test method to the
    marker IGNORE, while the child scenario adds the test method to the marker SKIP. This should throw an error on
    collecting level.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except ValueError as exc:
        assert exc.args[0] == ("the element `test_add_two_numbers` given at class attribute `ScenarioAddingChild.SKIP` "
                               "was already added to IGNORE in a higher parent class - not possible to add it now "
                               "to SKIP")

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('../../src'))

from balder import __version__


# -- Project information -----------------------------------------------------

project = 'balder'
version = __version__
# only use major/minor here
release = ".".join(version.split(".")[:2])

copyright = '2022-2025, Max Stahlschmidt'
author = 'Max Stahlschmidt'


# -- General configuration ---------------------------------------------------

html_context = {}

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinxcontrib.mermaid'
]

# output format for graphviz extension
graphviz_output_format = "svg"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

# Tell Jinja2 templates the build is running on Read the Docs
if os.environ.get("READTHEDOCS", "") == "True":
    html_context["READTHEDOCS"] = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = "<release>"
html_logo = "_static/balder.png"
html_theme_options = {
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = ['custom.css']

# -- mermaid

mermaid_version = ""

html_js_files = [
   'js/mermaid.js',
]



from multiprocessing import Process
from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_reverse_routing(balder_working_dir):
    """
    This testcase has an environment, that should be able to be executed in two possible variations. The only key
    difference to the basic env `test_calculator` is the reverse `@connect(..)` marker. In scenario, the provider devices
    are decorated. In setup, the calculator device is decorated with both `@connect(..)` marker
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"



import balder
import logging
from .setup_features import SetupFeatureI, SetupFeatureII
from ..lib.connections import AConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupAFactory:

    def get_setup(self):
        class SetupA(balder.Setup):
            """This is a setup of category A (exactly the same as scenario A)"""

            class SetupDevice1(balder.Device):
                s_i = SetupFeatureI()

            @balder.connect(SetupDevice1, over_connection=AConnection)
            class SetupDevice2(balder.Device):
                s_ii = SetupFeatureII()

            @balder.fixture(level="session")
            def fixture_session(self):
                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

            @balder.fixture(level="setup")
            def fixture_setup(self):
                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

            @balder.fixture(level="scenario")
            def fixture_scenario(self):
                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

            @balder.fixture(level="variation")
            def fixture_variation(self):
                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

            @balder.fixture(level="testcase")
            def fixture_testcase(self):
                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.SetupDevice1.s_i.do_something()
                self.SetupDevice2.s_ii.do_something()

        return SetupA

factory = SetupAFactory()

SetupA1 = factory.get_setup()



import traceback
from multiprocessing import Process
from _balder.balder_session import BalderSession
from balder.exceptions import IllegalConnectionTypeError


def test_0_use_custom_globtree(balder_working_dir):
    """
    This test tries to use a custom glob-tree, which is defined in the `BalderSettings` of the related environment. The
    test will be executed with some connections from the normal conn-tree. Balder has to throw an error, because all used
    connection are unknown for the newly created active conn-tree.
    """
    proc = Process(target=processed, args=(balder_working_dir,))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    exception = None
    try:
        session.run()
        print("\n")
    except IllegalConnectionTypeError as exc:
        traceback.print_exc()
        exception = exc
    assert exception, "no exception was thrown"



from __future__ import annotations

from _balder.exceptions import NodeNotExistsError
from _balder.controllers import DeviceController


class NodeGateway:
    """
    This is a class that describes a gateway between two nodes of a device. A gateway enables the interaction of
    different connection nodes of a device.
    """

    def __init__(self, device, from_node: str, to_node: str, bidirectional: bool):
        """

        :param device: the device this gateway belongs to

        :param from_node: the node name where the gateway starts

        :param to_node: the node name where the gateway ends

        :param bidirectional: if this value is true, the gateway can "translate" in both directions
        """
        #: the device this gateway belongs to
        self.device = device

        #: the node name of the device the gateway starts
        self.from_node_name = from_node

        #: the node name of the device the gateway ends
        self.to_node_name = to_node

        #: if this value is true, the gateway can "translate" in both directions
        self._bidirectional = bidirectional

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def validate_given_node_names(self):
        """
        This method validates whether the given node names actually exist in the transferred device. Otherwise, this
        method throws an exception.
        """
        device_controller = DeviceController.get_for(self.device)
        if self.from_node_name not in device_controller.get_all_connections().keys():
            raise NodeNotExistsError(f"the from_node `{self.from_node_name}` mentioned by the `@gateway` decorator "
                                     f"does not exist as a node in device `{self.__class__.__name__}`")
        if self.to_node_name not in device_controller.get_all_connections().keys():
            raise NodeNotExistsError(f"the to_node `{self.to_node_name}` mentioned by the `@gateway` decorator does "
                                     f"not exist as a node in device `{self.__class__.__name__}`")

    def get_conn_partner_of(self, node) -> str:
        """
        This method returns the partner node of this gateway - it always returns the other not given side
        """
        if node not in (self.from_node_name, self.to_node_name):
            raise ValueError(f"the given node `{node}` is no component of this connection")

        return self.to_node_name if node == self.from_node_name else self.from_node_name

    def has_connection_from_to(self, start_node, end_node=None) -> bool:
        """
        This method checks if there is a connection from ``start_node`` to ``end_node``. This will always return true
        if the gateway supports bidirectional communication or the ``start_node`` and ``end_node`` given in this method
        are also the ``start_node`` and ``end_node`` mentioned in this connection object.

        .. note::
            Of course the given node names have to be the nodes mentioned internally!

        :param start_node: the node the connection should start

        :param end_node: the node the connection should end (optional, will be autofilled automatically)

        :return: returns true if the given direction is possible
        """
        if start_node not in (self.from_node_name, self.to_node_name):
            raise ValueError(
                f"the given start_node `{start_node.__qualname__}` is no component of this connection")
        if end_node is not None:
            if end_node not in (self.from_node_name, self.to_node_name):
                raise ValueError(
                    f"the given end_node `{end_node.__qualname__}` is no component of this connection")
        else:
            end_node = self.from_node_name if start_node == self.to_node_name else self.to_node_name

        if self.is_bidirectional():
            return True

        if start_node == self.from_node_name and end_node == self.to_node_name:
            return True

        return False

    def is_bidirectional(self) -> bool:
        """returns true if the gateway works in both directional"""
        return self._bidirectional



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_01_envtester_methvar_class import Base01EnvtesterMethvarClass


class Test01EnvtesterMethvar(Base01EnvtesterMethvarClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS.

    It uses a reduced version of the general envtester environment. For this it uses the feature `MethVarFeature`, that
    is implemented on four levels. The first definition (`BaseMethVarFeature`) is as pre scenario feature. Its child
    class (the `ScenarioMethVarFeature`) is instantiated in the scenario `ScenarioA`. The child class of the
    `ScenarioMethVarFeature` is the `BetweenMethVarFeature`, which is not directly instantiated. The last child class is
    the `SetupMethVarFeature`, which is instantiated in all setup classes.

    This basic testcase has three different method variations, that are implemented on all four levels. The
    class-based-decorator allows three different VDevices (all have another Feature). The features only has one method
    `do_something_as_var()`, they provide an implementation for. All decorator for all vDevices are for the default
    :meth:`Connection` object.

    The environment provides three similar setup classes. All of them assign another VDevice:

    * `SetupI` assigns `VDeviceI` for `SetupI.SetupDevice2`
    * `SetupII` assigns `VDeviceII` for `SetupII.SetupDevice2`
    * `SetupIII` assigns `VDeviceIII` for `SetupIII.SetupDevice2`

    The environment checks in its test method `ScenarioA.test_a_1()` and in the TESTCASE and VARIATION level fixtures if
    the method variation is being called correctly.

    The test secures that the correct method variations will be called and the test session terminated with SUCCESS.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},

            [
                (
                    # FIXTURE-CONSTRUCTION: SetupI.fixture_session
                    {"cls": "SetupI", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupII.fixture_session
                    {"cls": "SetupII", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupIII.fixture_session
                    {"cls": "SetupIII", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                )
            ],

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            [
                self._generate_expected_data_from_setup_on("SetupI", "VDeviceI", "VDeviceI", "VDeviceI", "VDeviceI"),
                self._generate_expected_data_from_setup_on("SetupII", "VDeviceII", "VDeviceII", "VDeviceII",
                                                           "VDeviceII"),
                self._generate_expected_data_from_setup_on("SetupIII", "VDeviceIII", "VDeviceIII", "VDeviceIII",
                                                           "VDeviceIII"),
            ],

            # FIXTURE-TEARDOWN: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            [
                (
                    # FIXTURE-TEARDOWN: SetupI.fixture_session
                    {"cls": "SetupI", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupII.fixture_session
                    {"cls": "SetupII", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIII.fixture_session
                    {"cls": "SetupIII", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                )
            ],

            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere)
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS

    def _generate_expected_data_from_setup_on(
            self, setup_class_name, base_feature_identifier, scenario_feature_identifier,
            between_feature_identifier, setup_feature_identifier):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_setup
            {"cls": setup_class_name, "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_scenario
            {"cls": setup_class_name, "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_variation
            {"cls": setup_class_name, "meth": "fixture_variation", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_testcase
            {"cls": setup_class_name, "meth": "fixture_testcase", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # TESTCASE: ScenarioA.test_a_1
            {"cls": "ScenarioA", "meth": "test_a_1"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: SetupI.fixture_testcase
            {"cls": setup_class_name, "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: SetupI.fixture_variation
            {"cls": setup_class_name, "meth": "fixture_variation", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: SetupI.fixture_scenario
            {"cls": setup_class_name, "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: SetupI.fixture_setup
            {"cls": setup_class_name, "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
        )



from ...lib.features import FeatureII
from ...balderglob import RuntimeObserver


class SetupFeatureII2(FeatureII):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII2, SetupFeatureII2.do_something, "enter `SetupFeatureIIB.do_something`",
            category="feature")




from __future__ import annotations
import inspect
from _balder.device import Device


class InnerDeviceManagingMetaclass(type):
    """metaclass for all classes that holds :class:`Device` objects - sets the reference to the outer class for them"""
    def __new__(mcs, name, parents, dct):
        cls = super(InnerDeviceManagingMetaclass, mcs).__new__(mcs, name, parents, dct)
        for inner_item in dct.values():
            if inspect.isclass(inner_item) and issubclass(inner_item, Device):
                inner_item._outer_balder_class = cls

        return cls



from typing import Union
import balder


class ReceiveMessageFeature(balder.Feature):

    def start_listening(self) -> None:
        raise NotImplementedError("this method has to be implemented on setup level")

    def get_next_received_msg(self) -> Union[str, None]:
        raise NotImplementedError("this method has to be implemented on setup level")


class SendMessageFeature(balder.Feature):
    class Receiver(balder.VDevice):
        recv = ReceiveMessageFeature()

    def send(self, message: str) -> None:
        raise NotImplementedError("this method has to be implemented on setup level")



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0CollectOnlyAndShowDiscarded(Base0EnvtesterClass):
    """
    This testcase executes the basic ENV example with a specific command line argument
    ``--collect-only --show-discarded``.

    The test checks that no tests are executed and the printed information is like expected. It makes sure, that the
    ``--show-discarded`` argument is ignored.
    """

    @property
    def cmd_args(self):
        return ['--collect-only', '--show-discarded']

    @property
    def expected_data(self) -> tuple:
        return tuple()

    def validate_printed_output(self, stdout: str) -> bool:
        assert self._check_header_of_stdout(stdout, 2, 2), \
            f"problems within header output"
        assert len(stdout.splitlines()) == 5, f"detect more output lines (={len(stdout.splitlines())}) than expected"
        return True

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # make sure that  no executor tree was created
        assert session.executor_tree is None, f"executor tree is not None"



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0NoMatchBecausePatternSwitched(Base0EnvtesterClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS. It calls the balder
    environment with the following filter patterns for setup and for scenarios and validates that the correct tests are
    executed:

    `balder .. --only-with-scenario setups/setup*.py --only-with-setup scenarios/scenario_*.py`

    No elements should be executed here.
    """
    @property
    def cmd_args(self):
        return ["--only-with-scenario", "setups/setup*.py", "--only-with-setup", "scenarios/scenario_*.py"]

    @property
    def expected_data(self):
        return ()

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
            "test session does not terminates with NOT RUN"



from typing import Union
from queue import Queue
from ..lib.features import SendMessageFeature, ReceiveMessageFeature


class ReceiveOverLocalVarFeature(ReceiveMessageFeature):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.msg_buffer = []

    def start_listening(self) -> None:
        self.msg_buffer = []

    def get_next_received_msg(self) -> Union[str, None]:
        return self.msg_buffer.pop(0)


class SendOverLocalVarFeature(SendMessageFeature):

    class Receiver(SendMessageFeature.Receiver):
        recv = ReceiveOverLocalVarFeature()

    def send(self, message: str) -> None:
        self.Receiver.recv.msg_buffer.append(message)


class ReceiveOverQueueFeature(ReceiveMessageFeature):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.queue = Queue()

    def start_listening(self) -> None:
        self.queue = Queue()

    def get_next_received_msg(self) -> Union[str, None]:
        if self.queue.empty():
            return None
        return self.queue.get_nowait()


class SendOverQueueFeature(SendMessageFeature):
    class Receiver(SendMessageFeature.Receiver):
        recv = ReceiveOverQueueFeature()

    def send(self, message: str) -> None:
        self.Receiver.recv.queue.put(message)



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_no_setup_devices(balder_working_dir):
    """
    This testcase executes the CALCULATOR environment in a modified version. In this version the setup has no devices.
    Balder should run successful, but no variation should be found.

    This test secures that Balder terminates successful, but without a possible variation.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
        "test session does not terminates with result NOT_RUN"

    assert len(session.executor_tree.all_child_executors) == 0, "the executor tree is not empty"



from multiprocessing import Process

from _balder.balder_session import BalderSession


def test_1_covered_by_non_testmethod(balder_working_dir):
    """
    This testcase executes the basic example but uses a `covered_by` decorator with a non-testmethod.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except TypeError as exc:
        assert exc.args[0] == ("the given element for `item` must be a test method of a scenario class (has to start "
                               "with `test_`)")

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



import balder
from balder.parametrization import Parameter
import logging
from ..lib.features import FeatureIII, FeatureIV
from ..lib.connections import BConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class ScenarioB(balder.Scenario):
    """This is the scenario of category B"""

    class ScenarioDevice3(balder.Device):
        iii = FeatureIII()

    @balder.connect(ScenarioDevice3, over_connection=BConnection)
    class ScenarioDevice4(balder.Device):
        iv = FeatureIV()

    @balder.parametrize_by_feature('test', (ScenarioDevice3, 'iii', 'params_from_property'))
    def test_b_1(self, test):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.test_b_1, category="testcase",
                                  msg=f"execute Test `{ScenarioB.test_b_1.__qualname__}` with Parameter `test={test}`")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.parametrize_by_feature('one_multiple', (ScenarioDevice4, 'iv', 'get_next_three_multiple_of'), parameter={'value': Parameter('number')})
    @balder.parametrize_by_feature('number', (ScenarioDevice3, 'iii', 'params_from_property'))
    def test_b_2(self, one_multiple, number):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.test_b_2, category="testcase",
                                  msg=f"execute Test `{ScenarioB.test_b_2.__qualname__}` with Parameter "
                                      f"`number={number}` and `one_multiple={one_multiple}`")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        assert one_multiple % number == 0

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioB, ScenarioB.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()



from __future__ import annotations
from typing import Literal

import functools
from _balder.collector import Collector
from _balder.fixture_execution_level import FixtureExecutionLevel


def fixture(level: Literal['session', 'setup', 'scenario', 'variation', 'testcase']):
    """
    This decorator declares the decorated function/method as a fixture function/method.

    :param level: the execution level the fixture should have
    """
    allowed_levels = [level.value for level in FixtureExecutionLevel]

    if level not in allowed_levels:
        raise ValueError(f"the value of `level` must be a `str` with one of the values `{'`, `'.join(allowed_levels)}`")

    def decorator_fixture(func):
        # always register the raw fixture in Collector - class determination will be done later by :meth:`Collector`
        Collector.register_raw_fixture(func, level)

        @functools.wraps(func)
        def wrapper_fixture(*args, **kwargs):
            func(*args, **kwargs)

        return wrapper_fixture
    return decorator_fixture



from __future__ import annotations
from typing import List, Union, Type, Dict, TypeVar, TYPE_CHECKING
from abc import ABC, abstractmethod
from _balder.utils.functions import cnn_type_check_and_convert

if TYPE_CHECKING:
    from ..connection import Connection
    from ..connection import ConnectionMetadata
    from ..device import Device
    from .and_connection_relation import AndConnectionRelation
    from .or_connection_relation import OrConnectionRelation

BaseConnectionRelationT = TypeVar('BaseConnectionRelationT', bound='BaseConnectionRelation')


class BaseConnectionRelation(ABC):
    """
    This is the base connection list class, which provides the general functionality of connection collections.
    """

    def __init__(self, *connections: Union[Type[Connection], Connection, BaseConnectionRelationT]):
        self._connections = []
        # add it over append for type checking
        for connection in connections:
            self.append(connection)

    def __iter__(self):
        return self._connections.__iter__()

    def __getitem__(self, item):
        return self._connections.__getitem__(item)

    def __len__(self):
        return len(self._connections)

    def __hash__(self):
        all_hashes = 0
        for cur_elem in self.connections:
            all_hashes += hash(cur_elem) + hash(self.__class__.__name__)
        return all_hashes

    def __and__(
            self,
            other: Union[Connection, Type[Connection], AndConnectionRelation, OrConnectionRelation]
    ) -> AndConnectionRelation:

        from .and_connection_relation import AndConnectionRelation  # pylint: disable=import-outside-toplevel

        new_list = AndConnectionRelation()
        new_list.append(self)
        new_list.append(cnn_type_check_and_convert(other))
        return new_list

    def __or__(
            self,
            other: Union[Connection, Type[Connection], AndConnectionRelation, OrConnectionRelation]
    ) -> OrConnectionRelation:

        from .or_connection_relation import OrConnectionRelation  # pylint: disable=import-outside-toplevel

        new_list = OrConnectionRelation()
        new_list.append(self)
        new_list.append(cnn_type_check_and_convert(other))
        return new_list

    @property
    def connections(self) -> List[Union[Connection, BaseConnectionRelationT]]:
        """
        returns the components of this connection relation
        """
        return self._connections.copy()

    @property
    def metadata(self) -> ConnectionMetadata | None:
        """
        returns the metadata of this connection relation
        """
        if not self.connections:
            return None

        # get all unique metadata objects
        existing_metadata = list({elem.metadata for elem in self.connections if elem})
        if len(existing_metadata) > 1:
            raise ValueError(f'different metadata detected: `{existing_metadata}`')
        return existing_metadata[0]

    @abstractmethod
    def get_simplified_relation(self) -> OrConnectionRelation:
        """
        This method simplifies the connection relation. It will convert every possible relation into an
        OrConnectionRelation[Connection, AndConnectionRelation].
        """

    def clone(self):
        """
        clones this connection relation
        """
        return self.__class__(*[cnn.clone() for cnn in self._connections])

    def append(self, connection: Union[Type[Connection], Connection, BaseConnectionRelationT]):
        """
        appends a component to this connection relation
        """
        from ..connection import Connection  # pylint: disable=import-outside-toplevel

        if isinstance(connection, type):
            if issubclass(connection, Connection):
                connection = connection()
            else:
                raise TypeError(f'can not append a element `{connection.__name__}`')
        if not isinstance(connection, Connection) and not isinstance(connection, BaseConnectionRelation):
            raise TypeError('the element that should be appended to the relation needs to be a Connection or another '
                            'relation')
        if isinstance(connection, self.__class__):
            # directly add children (because it has the same type)
            for cur_inner_connection in connection.connections:
                self._connections.append(cur_inner_connection)
        else:
            self._connections.append(connection)

    def extend(self, relation: BaseConnectionRelationT):
        """
        extends the relation with the elements of provided relation
        """
        if not isinstance(relation, self.__class__):
            raise TypeError(f'the relation with the elements that should be appended needs to be from same type (is '
                            f'`{relation.__class__}` | expected `{self.__class__}`)')
        for cur_elem in relation.connections:
            self.append(cur_elem)

    def set_metadata_for_all_subitems(self, metadata: Union[Dict[str, Union[Device, str]], None]):
        """
        This method sets the metadata for all existing Connection items in this element.

        :param metadata: the metadata that should be set (if the value is explicitly set to None, it removes the
                         metadata from every item)
        """
        for cur_cnn in self._connections:
            cur_cnn.set_metadata_for_all_subitems(metadata)

    def get_all_used_connection_types(self) -> List[Type[Connection]]:
        """
        This method returns all available connection types, that are used within this connection relation.
        """
        from ..connection import Connection  # pylint: disable=import-outside-toplevel

        result = []
        for cur_inner_elem in self._connections:
            if isinstance(cur_inner_elem, Connection):
                result.append(cur_inner_elem.__class__)
            elif isinstance(cur_inner_elem, BaseConnectionRelation):
                result.extend(cur_inner_elem.get_all_used_connection_types())
            else:
                raise TypeError(f'unexpected type for inner item `{cur_inner_elem.__class__.__name__}`')
        return list(set(result))

    @abstractmethod
    def get_tree_str(self) -> str:
        """
        returns the tree string for this part of the connection tree
        """

    def is_resolved(self) -> bool:
        """
        returns whether this connection relation is part of a single connection
        """
        if len(self._connections) == 0:
            return True
        return min(cnn.is_resolved() for cnn in self._connections)

    def get_resolved(self) -> BaseConnectionRelationT:
        """
        This method returns a resolved Connection Tree. This means that it convert the based_on references so that
        every based on connection is a direct parent of the current element. It secures that there are no undefined
        connection layers between an object and the given parent.
        """
        return self.__class__(*[cnn.get_resolved() for cnn in self._connections])

    @abstractmethod
    def is_single(self) -> bool:
        """
        returns whether this connection relation is part of a single connection
        """

    @abstractmethod
    def get_singles(self) -> List[Connection]:
        """
        returns the single connections of all components of this connection relation
        """

    def cnn_are_in_other(self, other: BaseConnectionRelationT, ignore_metadata: bool=False) -> bool:
        """
        This method validates that the elements from this relation are contained in the other relation. Elements matches
        with each other, as soon as they are equal.

        .. note::
            This method only checks that every single connections from the first element is contained in the second too.
            It does not check the other direction. If you want to validate this, you need to call this method with both
            possibilities.

        :param other: the first list of connections
        :param other: the other relation object
        :param ignore_metadata: True, if the metadata of the single connections should be ignored
        :return: True in case that every connection of the first list is equal with one in the second list, otherwise
                 False
        """
        for cur_cnn in self._connections:
            found_equal = False
            for cur_other_cnn in other.connections:
                if cur_cnn.equal_with(cur_other_cnn, ignore_metadata=ignore_metadata):
                    found_equal = True
                    break
            if not found_equal:
                return False
        return True

    @abstractmethod
    def cut_into_all_possible_subtree_branches(self):
        """
        This method returns a list of all possible connection tree branches. A branch is a single connection, while
        this method returns a list of all possible singles where every single connection has this connection as head.
        """

    def equal_with(self, other_relation: BaseConnectionRelationT, ignore_metadata=False) -> bool:
        """
        This method returns True if the current relation matches with the other relation object. It always converts the
        elements to a resolved version and checks if both of them are exactly the same.

        .. note::
            Note that both connection relations need to be resolved.

        .. note::
            Note that both Connection objects have to have the same ending parent elements. Only the order of the inner
            connection elements are irrelevant.

        :param other_relation: the other connection relation (needs to be the same type)

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: returns True if both elements are same
        """
        if self.__class__ != other_relation.__class__:
            return False

        if not self.is_resolved():
            raise ValueError('can not execute method, because connection relation is not resolved')
        if not other_relation.is_resolved():
            raise ValueError('can not execute method, because other connection relation is not resolved')

        # check inner connection elements (if they match all in both directions)
        return (self.cnn_are_in_other(other_relation, ignore_metadata=ignore_metadata)
                and other_relation.cnn_are_in_other(self, ignore_metadata=ignore_metadata))

    @abstractmethod
    def contained_in(self, other_conn: Union[Connection, BaseConnectionRelationT], ignore_metadata=False) -> bool:
        """
        This method helps to find out whether this connection relation fits within a given connection tree. A connection
        object is a certain part of the large connection tree that Balder has at its disposal. This method checks
        whether a possibility of this connection tree fits in one possibility of the given connection tree.

        .. note::
            The method returns true if one single connection of this object fits in another single connection that is
            given by `other_conn`.

        :param other_conn: the other connection

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: true if the self object is contained in the `other_conn`, otherwise false
        """



import balder
from ..balderglob import RuntimeObserver


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")


class FeatureIOverwritten(FeatureI):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIOverwritten, FeatureIOverwritten.do_something, "enter `FeatureIChild.do_something`",
            category="feature")


class FeatureII(balder.Feature):

    class Dev1(balder.VDevice):
        feat = FeatureI()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")


class NewlyDefinedFeature(balder.Feature):
    """this feature will be instantiated in `ScenarioAChild` (and not in `ScenarioAParent`)"""
    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, NewlyDefinedFeature, NewlyDefinedFeature.do_something, "enter `NewlyDefinedFeature.do_something`",
            category="feature")


class FeatureIII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIII, FeatureIII.do_something, "enter `FeatureIII.do_something`", category="feature")


class FeatureIV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.do_something, "enter `FeatureIV.do_something`", category="feature")



import balder
import logging
from ..lib.features import FeatureIII, FeatureIV
from ..lib.connections import BConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class ScenarioBParent(balder.Scenario):
    """This is the scenario of category B"""

    class ScenarioDevice3(balder.Device):
        iii = FeatureIII()

    @balder.connect(ScenarioDevice3, over_connection=BConnection)
    class ScenarioDevice4(balder.Device):
        iv = FeatureIV()

    def test_b_1(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.test_b_1, category="testcase",
                                  msg=f"execute Test `{ScenarioBParent.test_b_1.__qualname__}`")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    def test_b_2(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.test_b_2, category="testcase",
                                  msg=f"execute Test `{ScenarioBParent.test_b_2.__qualname__}`")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioBParent, ScenarioBParent.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()
